@@ -147,12 +147,12 @@ public class Stub {
 		Assembly assembly = null;
 		try {
 			assembly = Assembly.LoadFile (name);
-		} catch (System.IO.FileNotFoundException e) { }
+		} catch (System.IO.FileNotFoundException) { }
 
 		if (assembly == null) {
 			try {
 				assembly = Assembly.LoadWithPartialName (name);
-			} catch (Exception e) { }
+			} catch (Exception) { }
 		}
 			
 		if (assembly == null)
@@ -402,12 +402,17 @@ public class Stub {
 				nsnode.SetAttribute("Name", type.Namespace);
 				index_types.AppendChild(nsnode);
 			}
+			string doc_typename = GetDocTypeName (type);
 			XmlElement typenode = (XmlElement)nsnode.SelectSingleNode("Type[@Name='" + typename + "']");
 			if (typenode == null) {
 				typenode = index_types.OwnerDocument.CreateElement("Type");
 				typenode.SetAttribute("Name", typename);
 				nsnode.AppendChild(typenode);
 			}
+			if (typename != doc_typename)
+				typenode.SetAttribute("DisplayName", doc_typename);
+			else
+				typenode.RemoveAttribute("DisplayName");
 				
 			// Ensure the namespace index file exists
 			if (!new System.IO.FileInfo(dest + "/" + type.Namespace + ".xml").Exists) {
@@ -427,8 +432,8 @@ public class Stub {
 			foreach (System.IO.FileInfo typefile in nsdir.GetFiles("*.xml")) {
 				if (!goodfiles.ContainsKey(nsdir.Name + "/" + typefile.Name)) {
 					string newname = typefile.FullName + ".remove";
-					try { System.IO.File.Delete(newname); } catch (Exception e) { }
-					try { typefile.MoveTo(newname); } catch (Exception e) { }					
+					try { System.IO.File.Delete(newname); } catch (Exception) { }
+					try { typefile.MoveTo(newname); } catch (Exception) { }					
 					Console.Error.WriteLine("Class no longer present; file renamed: " + nsdir.Name + "/" + typefile.Name);
 				}
 			}
@@ -458,10 +463,11 @@ public class Stub {
 		if (!ignoremembers)
 		foreach (XmlElement oldmember in basefile.SelectNodes("Type/Members/Member")) {
 			MemberInfo oldmember2 = GetMember(type, oldmember);
+ 			string sig = oldmember2 != null ? MakeMemberSignature(oldmember2) : null;
 			
 			// Interface implementations and overrides are deleted from the docs
 			// unless the overrides option is given.
-			if (oldmember2 != null && !IsNew(oldmember2))
+			if (oldmember2 != null && (!IsNew(oldmember2) || sig == null))
 				oldmember2 = null;
 			
 			// Deleted (or signature changed)
@@ -487,14 +493,17 @@ public class Stub {
 			// Update signature information
 			UpdateMember(oldmember, oldmember2);
 			
-			seenmembers[oldmember2] = 1;
+			seenmembers[sig] = 1;
 		}
 		
 		if (!IsDelegate(type) && !ignoremembers) {
 			XmlNode members = basefile.SelectSingleNode("Type/Members");
 			foreach (MemberInfo m in type.GetMembers(BindingFlags.Public|BindingFlags.NonPublic|BindingFlags.Static|BindingFlags.Instance|BindingFlags.DeclaredOnly)) {
 				if (m is Type) continue;
-				if (seenmembers.ContainsKey(m)) continue;
+				
+				string sig = MakeMemberSignature(m);
+				if (sig == null) continue;
+				if (seenmembers.ContainsKey(sig)) continue;
 				
 				// To be nice on diffs, members/properties/events that are overrides or are interface implementations
 				// are not added in.
@@ -585,9 +594,16 @@ public class Stub {
 		ArrayList memberparams = new ArrayList();
 		foreach (XmlElement param in member.SelectNodes("Parameters/Parameter"))
 			memberparams.Add(param.GetAttribute("Type"));
+
+		string memberName = member.GetAttribute ("MemberName");
+		do {
+			int idx = memberName.IndexOf ("<");
+			if (idx != -1)
+				memberName = memberName.Substring (0, idx);
+		} while (false);
 		
 		// Loop through all members in this type with the same name
-		MemberInfo[] mis = type.GetMember(member.GetAttribute("MemberName"), BindingFlags.Static | BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.DeclaredOnly);
+		MemberInfo[] mis = type.GetMember(memberName, BindingFlags.Static | BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.DeclaredOnly);
 		foreach (MemberInfo mi in mis) {
 			if (mi is Type) continue;
 			if (GetMemberType(mi) != membertype) continue;
@@ -597,7 +613,8 @@ public class Stub {
 
 			if (mi is MethodInfo) {
 				// Casting operators can overload based on return type.
-				if (returntype != ((MethodInfo)mi).ReturnType.FullName) continue;
+				if (returntype != GetDocTypeFullName (((MethodInfo)mi).ReturnType)) 
+					continue;
 			}
 
 			ParameterInfo[] pis = null;
@@ -613,7 +630,10 @@ public class Stub {
 			
 			bool good = true;
 			for (int i = 0; i < pis.Length; i++)
-				if (pis[i].ParameterType.FullName != (string)memberparams[i]) { good = false; break; }
+				if (GetDocParameterType (pis[i].ParameterType) != (string)memberparams[i]) {
+					good = false;
+					break;
+				}
 			if (!good) continue;
 
 			return mi;
@@ -628,12 +648,8 @@ public class Stub {
 		string typesig = MakeTypeSignature(type);
 		if (typesig == null) return null; // not publicly visible
 		
-		string typename = type.FullName.Substring(type.Namespace == "" ? 0 : type.Namespace.Length+1);
-
 		XmlDocument doc = new XmlDocument();
 		XmlElement root = doc.CreateElement("Type");
-		root.SetAttribute("Name", typename);
-		root.SetAttribute("FullName", type.FullName);
 		
 		UpdateType(root, type, true);
 		
@@ -643,6 +659,9 @@ public class Stub {
 	// STUBBING/UPDATING FUNCTIONS
 	
 	public static void UpdateType(XmlElement root, Type type, bool addmembers) {
+		root.SetAttribute("Name", GetDocTypeName (type));
+		root.SetAttribute("FullName", GetDocTypeFullName (type));
+
 		WriteElementAttribute(root, "TypeSignature[@Language='C#']", "Language", "C#");
 		WriteElementAttribute(root, "TypeSignature[@Language='C#']", "Value", MakeTypeSignature(type));
 		
@@ -664,10 +683,40 @@ public class Stub {
 		
 		NormalizeWhitespace(ass);
 		
+		if (type.IsGenericType) {
+			XmlElement typeargs = WriteElement(root, "TypeParameters");
+			ClearElementChildren(typeargs);
+			foreach (Type arg in type.GetGenericArguments()) {
+				XmlElement argnode = typeargs.OwnerDocument.CreateElement("TypeParameter");
+				typeargs.AppendChild(argnode);
+				argnode.InnerText = arg.Name;
+			}
+		} else {
+			ClearElement(root, "TypeParameters");
+		}
+		
 		if (type.BaseType != null) {
-			string basetypename = type.BaseType.FullName;
+			XmlElement basenode = WriteElement(root, "Base");
+			
+			string basetypename = GetDocTypeFullName (type.BaseType);
 			if (basetypename == "System.MulticastDelegate") basetypename = "System.Delegate";
 			WriteElementText(root, "Base/BaseTypeName", basetypename);
+			
+			// Document how this type instantiates the generic parameters of its base type
+			if (type.BaseType.IsGenericType) {
+				ClearElement(basenode, "BaseTypeArguments");
+				Type genericDefinition = type.BaseType.GetGenericTypeDefinition();
+				for (int i = 0; i < type.BaseType.GetGenericArguments().Length; i++) {
+					Type typearg = type.BaseType.GetGenericArguments()[i];
+					Type typeparam = genericDefinition.GetGenericArguments()[i];
+					
+					XmlElement bta = WriteElement(basenode, "BaseTypeArguments");
+					XmlElement arg = bta.OwnerDocument.CreateElement("BaseTypeArgument");
+					bta.AppendChild(arg);
+					arg.SetAttribute("TypeParamName", typeparam.Name);
+					arg.InnerText = GetDocTypeFullName (typearg);
+				}
+			}
 		} else {
 			ClearElement(root, "Base");
 		}
@@ -679,7 +728,7 @@ public class Stub {
 			ArrayList interface_names = new ArrayList();
 			foreach (Type i in type.GetInterfaces())
 				if ((type.BaseType == null || Array.IndexOf(type.BaseType.GetInterfaces(), i) == -1) && InterfaceNotFromAnother(i, type.GetInterfaces()))
-					interface_names.Add(i.FullName);
+					interface_names.Add(GetDocTypeFullName (i));
 			interface_names.Sort();
 
 			XmlElement interfaces = WriteElement(root, "Interfaces");
@@ -734,14 +783,23 @@ public class Stub {
 			WriteElementText(me, "MemberValue", fieldValue);
 		
 		ParameterInfo[] paraminfo = null;
+		Type[] genericParams = null;
 		Type returntype = null;
 		bool returnisreturn = true; // the name of the node is 'return', vs 'value'
 		bool addremarks = true;
 		
-		if (mi is MethodInfo || mi is ConstructorInfo)
+		if (mi is MethodInfo || mi is ConstructorInfo) {
 			paraminfo = ((MethodBase)mi).GetParameters();
-		else if (mi is PropertyInfo)
+			try {
+				genericParams = ((MethodBase)mi).GetGenericArguments ();
+			}
+			catch (NotSupportedException) {
+				genericParams = null;
+			}
+		}
+		else if (mi is PropertyInfo) {
 			paraminfo = ((PropertyInfo)mi).GetIndexParameters();
+		}
 			
 		if (mi is MethodInfo) {
 			returntype = ((MethodInfo)mi).ReturnType;
@@ -756,7 +814,7 @@ public class Stub {
 
 		string slashdocsig = MakeSlashDocMemberSig(mi);
 
-		MakeDocNode(me, paraminfo, returntype, returnisreturn, addremarks, slashdocsig);
+		MakeDocNode(me, genericParams, paraminfo, returntype, returnisreturn, addremarks, slashdocsig);
 		
 	}
 	
@@ -793,6 +851,8 @@ public class Stub {
 					ret = parent.OwnerDocument.CreateElement(ename);
 					parent.AppendChild(ret);
 					parent = ret;
+				} else {
+					parent = ret;
 				}
 			}
 		}
@@ -819,106 +879,59 @@ public class Stub {
 		if (node != null)
 			parent.RemoveChild(node);
 	}
+	private static void ClearElementChildren(XmlElement parent) {
+		while (parent.FirstChild != null)
+			parent.RemoveChild(parent.FirstChild);
+	}
 	
 	// DOCUMENTATION HELPER FUNCTIONS
 	
 	private static void MakeDocNode(Type type, XmlElement e, bool addremarks) {
 		ParameterInfo[] parameters = null;
+		Type[] genericParams = type.GetGenericArguments ();
+		if (type.DeclaringType != null) {
+			Type[] declGenParams = type.DeclaringType.GetGenericArguments ();
+			if (declGenParams != null && genericParams.Length == declGenParams.Length) {
+				genericParams = null;
+			}
+			else if (declGenParams != null) {
+				Type[] nestedParams = new Type [genericParams.Length - declGenParams.Length];
+				for (int i = 0; i < nestedParams.Length; ++i) {
+					nestedParams [i] = genericParams [i+declGenParams.Length];
+				}
+				genericParams = nestedParams;
+			}
+		}
 		Type returntype = null;
 		if (IsDelegate(type)) {
 			parameters = type.GetMethod("Invoke").GetParameters();
 			returntype = type.GetMethod("Invoke").ReturnType;
 		}
-		MakeDocNode(parameters, returntype, true, e, addremarks, "T:" + type.FullName);
+		MakeDocNode(genericParams, parameters, returntype, true, e, addremarks, "T:" + GetDocTypeFullName (type));
 	}
 	
-	private static void MakeDocNode(XmlElement root, ParameterInfo[] parameters, Type returntype, bool returnisreturn, bool addremarks, string slashdocsig) {
+	private static void MakeDocNode(XmlElement root, Type[] genericParams, ParameterInfo[] parameters, Type returntype, bool returnisreturn, bool addremarks, string slashdocsig) {
 		XmlElement e = WriteElement(root, "Docs");
-		MakeDocNode(parameters, returntype, returnisreturn, e, addremarks, slashdocsig);
+		MakeDocNode(genericParams, parameters, returntype, returnisreturn, e, addremarks, slashdocsig);
 	}
 	
-	private static void MakeDocNode(ParameterInfo[] parameters, Type returntype, bool returnisreturn, XmlElement e, bool addremarks, string slashdocname) {
+	private static void MakeDocNode(Type[] genericParams, ParameterInfo[] parameters, Type returntype, bool returnisreturn, XmlElement e, bool addremarks, string slashdocname) {
 		WriteElementInitialText(e, "summary", "To be added.");
 		
 		if (parameters != null) {
-			XmlNode[] paramnodes = new XmlNode[parameters.Length];
-			
-			// Some documentation had param nodes with leading spaces.
-			foreach (XmlElement paramnode in e.SelectNodes("param")){
-				paramnode.SetAttribute("name", paramnode.GetAttribute("name").Trim());
-			}
-			
-			// If a member has only one parameter, we can track changes to
-			// the name of the parameter easily.
-			if (parameters.Length == 1) {
-				int c = e.SelectNodes("param").Count;
-				if (c == 1) {
-					UpdateParameterName (e, (XmlElement) e.SelectSingleNode("param"), parameters[0].Name);
-				} else if (c > 1) {
-					foreach (XmlElement paramnode in e.SelectNodes("param")) {
-						if (paramnode.InnerText.StartsWith("To be added"))
-							paramnode.ParentNode.RemoveChild(paramnode);
-						else
-							UpdateParameterName (e, paramnode, parameters[0].Name);
-					}
-				}
-			}
-
-			bool reinsert = false;
-
-			// Pick out existing and still-valid param nodes, and
-			// create nodes for parameters not in the file.
-			Hashtable seenParams = new Hashtable();
-			for (int pi = 0; pi < parameters.Length; pi++) {
-				ParameterInfo p = parameters[pi];
-				seenParams[p.Name] = pi;
-				
-				paramnodes[pi] = e.SelectSingleNode("param[@name='" + p.Name + "']");
-				if (paramnodes[pi] != null) continue;
-				
-				XmlElement pe = e.OwnerDocument.CreateElement("param");
-				pe.SetAttribute("name", p.Name);
-				pe.InnerText = "To be added.";
-				paramnodes[pi] = pe;
-				reinsert = true;
-			}
-
-			// Remove parameters that no longer exist and check all params are in the right order.
-			int idx = 0;
-			foreach (XmlElement paramnode in e.SelectNodes("param")) {
-				string name = paramnode.GetAttribute("name");
-				if (!seenParams.ContainsKey(name)) {
-					if (!delete && !paramnode.InnerText.StartsWith("To be added")) {
-						Console.Error.Write("The following param node can only be deleted if the --delete option is given: ");
-						Console.Error.WriteLine(paramnode.OuterXml);
-					} else {
-						paramnode.ParentNode.RemoveChild(paramnode);
-					}
-					continue;
-				}
-
-				if ((int)seenParams[name] != idx)
-					reinsert = true;
-				
-				idx++;
-			}
-			
-			// Re-insert the parameter nodes at the top of the doc section.
-			if (reinsert)
-				for (int pi = parameters.Length-1; pi >= 0; pi--)
-					e.PrependChild(paramnodes[pi]);
-		} else {
-			// Clear all existing param nodes
-			foreach (XmlNode paramnode in e.SelectNodes("param")) {
-				if (!delete && !paramnode.InnerText.StartsWith("To be added")) {
-					Console.Error.WriteLine("The following param node can only be deleted if the --delete option is given:");
-					Console.Error.WriteLine(paramnode.OuterXml);
-				} else {
-					paramnode.ParentNode.RemoveChild(paramnode);
-				}
-			}
+			string[] values = new string [parameters.Length];
+			for (int i = 0; i < values.Length; ++i)
+				values [i] = parameters [i].Name;
+			UpdateParameters (e, "param", values);
 		}
-		
+
+		if (genericParams != null) {
+			string[] values = new string [genericParams.Length];
+			for (int i = 0; i < values.Length; ++i)
+				values [i] = genericParams [i].Name;
+			UpdateParameters (e, "typeparam", values);
+		}
+
 		string retnodename = null;
 		if (returntype != null && returntype != typeof(void)) {
 			retnodename = returnisreturn ? "returns" : "value";
@@ -972,6 +985,88 @@ public class Stub {
 		NormalizeWhitespace(e);
 	}
 
+	private static void UpdateParameters (XmlElement e, string element, string[] values)
+	{	
+		if (values != null) {
+			XmlNode[] paramnodes = new XmlNode[values.Length];
+			
+			// Some documentation had param nodes with leading spaces.
+			foreach (XmlElement paramnode in e.SelectNodes(element)){
+				paramnode.SetAttribute("name", paramnode.GetAttribute("name").Trim());
+			}
+			
+			// If a member has only one parameter, we can track changes to
+			// the name of the parameter easily.
+			if (values.Length == 1) {
+				int c = e.SelectNodes(element).Count;
+				if (c == 1) {
+					UpdateParameterName (e, (XmlElement) e.SelectSingleNode(element), values [0]);
+				} else if (c > 1) {
+					foreach (XmlElement paramnode in e.SelectNodes(element)) {
+						if (paramnode.InnerText.StartsWith("To be added"))
+							paramnode.ParentNode.RemoveChild(paramnode);
+						else
+							UpdateParameterName (e, paramnode, values [0]);
+					}
+				}
+			}
+
+			bool reinsert = false;
+
+			// Pick out existing and still-valid param nodes, and
+			// create nodes for parameters not in the file.
+			Hashtable seenParams = new Hashtable();
+			for (int pi = 0; pi < values.Length; pi++) {
+				string p = values [pi];
+				seenParams[p] = pi;
+				
+				paramnodes[pi] = e.SelectSingleNode(element + "[@name='" + p + "']");
+				if (paramnodes[pi] != null) continue;
+				
+				XmlElement pe = e.OwnerDocument.CreateElement(element);
+				pe.SetAttribute("name", p);
+				pe.InnerText = "To be added.";
+				paramnodes[pi] = pe;
+				reinsert = true;
+			}
+
+			// Remove parameters that no longer exist and check all params are in the right order.
+			int idx = 0;
+			foreach (XmlElement paramnode in e.SelectNodes(element)) {
+				string name = paramnode.GetAttribute("name");
+				if (!seenParams.ContainsKey(name)) {
+					if (!delete && !paramnode.InnerText.StartsWith("To be added")) {
+						Console.Error.Write("The following param node can only be deleted if the --delete option is given: ");
+						Console.Error.WriteLine(paramnode.OuterXml);
+					} else {
+						paramnode.ParentNode.RemoveChild(paramnode);
+					}
+					continue;
+				}
+
+				if ((int)seenParams[name] != idx)
+					reinsert = true;
+				
+				idx++;
+			}
+			
+			// Re-insert the parameter nodes at the top of the doc section.
+			if (reinsert)
+				for (int pi = values.Length-1; pi >= 0; pi--)
+					e.PrependChild(paramnodes[pi]);
+		} else {
+			// Clear all existing param nodes
+			foreach (XmlNode paramnode in e.SelectNodes(element)) {
+				if (!delete && !paramnode.InnerText.StartsWith("To be added")) {
+					Console.Error.WriteLine("The following param node can only be deleted if the --delete option is given:");
+					Console.Error.WriteLine(paramnode.OuterXml);
+				} else {
+					paramnode.ParentNode.RemoveChild(paramnode);
+				}
+			}
+		}
+	}
+
 	private static void UpdateParameterName (XmlElement docs, XmlElement pe, string newName)
 	{
 		string existingName = pe.GetAttribute ("name");
@@ -994,9 +1089,20 @@ public class Stub {
 				n.ParentNode.RemoveChild(n);
 	}
 	
-	private static void MakeAttributes(XmlElement root, ICustomAttributeProvider attributes, bool assemblyAttributes) {
-		object[] at = attributes.GetCustomAttributes(false);
-		if (at.Length == 0) {
+	private static void MakeAttributes(XmlElement root, object attributes, bool assemblyAttributes) {
+		System.Collections.Generic.IList<CustomAttributeData> at;
+		if (attributes is Assembly)
+			at = CustomAttributeData.GetCustomAttributes((Assembly)attributes);
+		else if (attributes is MemberInfo)
+			at = CustomAttributeData.GetCustomAttributes((MemberInfo)attributes);
+		else if (attributes is Module)
+			at = CustomAttributeData.GetCustomAttributes((Module)attributes);
+		else if (attributes is ParameterInfo)
+			at = CustomAttributeData.GetCustomAttributes((ParameterInfo)attributes);
+		else
+			throw new ArgumentException("unsupported type: " + attributes.GetType().ToString());
+	
+		if (at.Count == 0) {
 			ClearElement(root, "Attributes");
 			return;
 		}
@@ -1008,36 +1114,30 @@ public class Stub {
 		else
 			e = root.OwnerDocument.CreateElement("Attributes");
 		
-		foreach (Attribute a in at) {
-			if (GetTypeVisibility(a.GetType().Attributes) == null) continue; // hide non-visible attributes
+		foreach (CustomAttributeData a in at) {
+			if (GetTypeVisibility(a.Constructor.DeclaringType.Attributes) == null) continue; // hide non-visible attributes
 			
-			//if (assemblyAttributes && a.GetType().FullName.StartsWith("System.Reflection.")) continue;
-			if (a.GetType().FullName == "System.Reflection.AssemblyKeyFileAttribute" || a.GetType().FullName == "System.Reflection.AssemblyDelaySignAttribute") continue; // hide security-related attributes
+			if (a.Constructor.DeclaringType == typeof(System.Reflection.AssemblyKeyFileAttribute) || a.Constructor.DeclaringType == typeof(System.Reflection.AssemblyDelaySignAttribute)) continue; // hide security-related attributes
+			if (a.Constructor.DeclaringType == typeof(System.Runtime.InteropServices.OutAttribute)) continue; // hide this because it is given in RefType attribute
 			
 			b = true;
 			
-			// There's no way to reconstruct how the attribute's constructor was called,
-			// so as a substitute, just list the value of all of the attribute's public fields.
-			
 			ArrayList fields = new ArrayList();
-			foreach (PropertyInfo f in a.GetType().GetProperties(BindingFlags.Public|BindingFlags.Instance)) {
-				if (f.Name == "TypeId") continue;
-				
-				object v = f.GetValue(a, null);
-				if (v == null) v = "null";
-				else if (v is string) v = "\"" + v + "\"";
-				else if (v is Type) v = "typeof(" + ((Type)v).FullName + ")";
-				else if (v is Enum) v = v.GetType().FullName + "." + v.ToString().Replace(", ", "|");
-					
-				fields.Add(f.Name + "=" + v);
+
+			foreach (CustomAttributeTypedArgument f in a.ConstructorArguments) {
+				fields.Add(MakeAttributesValueString(f.Value));
 			}
+			foreach (CustomAttributeNamedArgument f in a.NamedArguments) {
+				fields.Add(f.MemberInfo.Name + "=" + MakeAttributesValueString(f.TypedValue.Value));
+			}
+
 			string a2 = String.Join(", ", (string[])fields.ToArray(typeof(string)));
 			if (a2 != "") a2 = "(" + a2 + ")";
 			
 			XmlElement ae = root.OwnerDocument.CreateElement("Attribute");
 			e.AppendChild(ae);
 			
-			string name = a.GetType().FullName;
+			string name = a.Constructor.DeclaringType.FullName;
 			if (name.EndsWith("Attribute")) name = name.Substring(0, name.Length-"Attribute".Length);
 			WriteElementText(ae, "AttributeName", name + a2);
 		}
@@ -1050,6 +1150,15 @@ public class Stub {
 		NormalizeWhitespace(e);
 	}
 	
+	private static string MakeAttributesValueString(object v) {
+		if (v == null) return "null";
+		else if (v is string) return "\"" + v + "\"";
+		else if (v is bool) return (bool)v ? "true" : "false";
+		else if (v is Type) return "typeof(" + GetCSharpFullName ((Type)v) + ")";
+		else if (v is Enum) return v.GetType().FullName + "." + v.ToString().Replace(", ", "|");
+		else return v.ToString();
+	}
+	
 	private static void MakeParameters(XmlElement root, ParameterInfo[] parameters) {
 		XmlElement e = WriteElement(root, "Parameters");
 		e.RemoveAll();
@@ -1057,7 +1166,7 @@ public class Stub {
 			XmlElement pe = root.OwnerDocument.CreateElement("Parameter");
 			e.AppendChild(pe);
 			pe.SetAttribute("Name", p.Name);
-			pe.SetAttribute("Type", p.ParameterType.FullName);
+			pe.SetAttribute("Type", GetDocParameterType (p.ParameterType));
 			if (p.ParameterType.IsByRef) {
 				if (p.IsOut) pe.SetAttribute("RefType", "out");
 				else pe.SetAttribute("RefType", "ref");
@@ -1081,10 +1190,15 @@ public class Stub {
 		else throw new ArgumentException();
 	}
 
+	private static string GetDocParameterType (Type type)
+	{
+		return GetDocTypeFullName (type).Replace ("@", "&");
+	}
+
 	private static void MakeReturnValue(XmlElement root, Type type, ICustomAttributeProvider attributes) {
 		XmlElement e = WriteElement(root, "ReturnValue");
 		e.RemoveAll();
-		WriteElementText(e, "ReturnType", type.FullName);
+		WriteElementText(e, "ReturnType", GetDocTypeFullName (type));
 		if (attributes != null)
 			MakeAttributes(root, attributes, false);
 	}
@@ -1126,12 +1240,35 @@ public class Stub {
 		if (mi.Name.StartsWith("raise_")) return null;
 		
 		XmlElement me = doc.CreateElement("Member");
-		me.SetAttribute("MemberName", mi.Name);
+		me.SetAttribute("MemberName", GetMemberName (mi));
 		
 		UpdateMember(me, mi);
 		
 		return me;
 	}
+
+	private static string GetMemberName (MemberInfo mi)
+	{
+		MethodBase mb = mi as MethodBase;
+		if (mb == null)
+			return mi.Name;
+		try {
+			StringBuilder sb = new StringBuilder ();
+			sb.Append (mi.Name);
+			Type[] typeParams = mb.GetGenericArguments ();
+			if (typeParams.Length > 0) {
+				sb.Append ("<");
+				sb.Append (typeParams [0].Name);
+				for (int i = 1; i < typeParams.Length; ++i)
+					sb.Append (",").Append (typeParams [i].Name);
+				sb.Append (">");
+			}
+			return sb.ToString ();
+		}
+		catch (NotSupportedException) {
+			return mi.Name;
+		}
+	}
 	
 	static bool IsDelegate(Type type) {
 		return typeof(System.Delegate).IsAssignableFrom (type) && !type.IsAbstract;
@@ -1184,10 +1321,10 @@ public class Stub {
 			sig.Append("delegate ");
 			MethodInfo invoke = type.GetMethod ("Invoke");
 			string arguments = GetMethodParameters(invoke.GetParameters());
-			string return_value = ConvertCTSName(invoke.ReturnType.FullName);
+			string return_value = GetCSharpFullName (invoke.ReturnType);
 			sig.Append(return_value);
 			sig.Append(" ");
-			sig.Append(type.Name);
+			sig.Append(GetCSharpName (type));
 			sig.Append("(");
 			sig.Append(arguments);
 			sig.Append(");");
@@ -1196,24 +1333,24 @@ public class Stub {
 		
 		sig.Append(GetTypeKind(type));
 		sig.Append(" ");
-		sig.Append(type.Name);
+		sig.Append(GetCSharpName (type));
 
-		if (!type.IsValueType && !type.IsEnum) {
+		if (!type.IsEnum) {
 			Type basetype = type.BaseType;
-			if (basetype == typeof(object)) // don't show this in signatures
+			if (basetype == typeof(object) || type.IsValueType) // don't show this in signatures
 				basetype = null;
 			
 			ArrayList interface_names = new ArrayList();
 			foreach (Type i in type.GetInterfaces())
 				if ((type.BaseType == null || Array.IndexOf(type.BaseType.GetInterfaces(), i) == -1) && InterfaceNotFromAnother(i, type.GetInterfaces()))
-					interface_names.Add(i.FullName);
+					interface_names.Add(GetCSharpFullName (i));
 			interface_names.Sort();
 			
 			if (basetype != null || interface_names.Count > 0)
 				sig.Append(" : ");
 			
 			if (basetype != null) {
-				sig.Append(basetype.FullName);
+				sig.Append(GetCSharpFullName (basetype));
 				if (interface_names.Count > 0)
 					sig.Append(", ");
 			}
@@ -1242,7 +1379,7 @@ public class Stub {
 		if (visibility == null) return null;
 		if (field.DeclaringType.IsEnum) return field.Name;
 		
-		string type = ConvertCTSName (field.FieldType.FullName);
+		string type = GetCSharpFullName (field.FieldType);
 		
 		string modifiers = String.Empty;
 		if (field.IsStatic && !field.IsLiteral) modifiers += " static";
@@ -1277,10 +1414,7 @@ public class Stub {
 				if (parameter.IsOut) sb.Append("out ");
 				else sb.Append("ref ");
 			}
-			string param = parameter.ParameterType.FullName;
-			if (parameter.ParameterType.IsByRef)
-				param = param.Substring (0, param.Length - 1);
-			param = ConvertCTSName(param);			
+			string param = GetCSharpFullName (parameter.ParameterType).Replace ("@", "");
 			sb.Append (param);
 			sb.Append (" ");
 			sb.Append (parameter.Name);
@@ -1291,8 +1425,9 @@ public class Stub {
 	}
 
 	static string MakeMethodSignature (MethodInfo method) {
-		string visibility = GetMethodVisibility (method);
-		if (visibility == null)
+		StringBuilder sb = new StringBuilder ();
+		sb.Append (GetMethodVisibility (method));
+		if (sb.Length == 0)
 				return null;
 		
 		string modifiers = String.Empty;
@@ -1305,19 +1440,30 @@ public class Stub {
 		if (method.IsFinal) modifiers += " sealed";
 		if (modifiers == " virtual sealed") modifiers = "";
 
+		sb.Append (modifiers);
+
 		// Special signature for destructors.
 		if (method.Name == "Finalize" && method.GetParameters().Length == 0)
 			return "~" + method.DeclaringType.Name + " ();";	
 
-		string return_type = ConvertCTSName (method.ReturnType.FullName);
-		string parameters = GetMethodParameters (method.GetParameters());
+		sb.Append (" ").Append (GetCSharpFullName (method.ReturnType));
+		sb.Append (" ").Append (method.Name);
+		try {
+			Type[] args = method.GetGenericArguments ();
+			if (args.Length > 0) {
+				sb.Append ("<");
+				sb.Append (args [0].Name);
+				for (int i = 1; i < args.Length; ++i)
+					sb.Append (",").Append (args [i].Name);
+				sb.Append (">");
+			}
+		}
+		catch (NotSupportedException) {
+			/* ignore */
+		}
+		sb.Append (" (").Append (GetMethodParameters (method.GetParameters())).Append (");");
 
-		string method_name = method.Name;
-		
-		// operators, default accessors need name rewriting
-
-		return String.Format ("{0}{1} {2} {3} ({4});",
-						visibility, modifiers, return_type, method_name, parameters);
+		return sb.ToString ();
 	}
 
 	static string MakeConstructorSignature (ConstructorInfo constructor) {
@@ -1325,7 +1471,7 @@ public class Stub {
 		if (visibility == null)
 			return null;
 
-		string name = constructor.DeclaringType.Name;
+		string name = GetTypeName (constructor.DeclaringType.Name);
 		string parameters = GetMethodParameters (constructor.GetParameters());
 
 		return String.Format ("{0} {1} ({2});",
@@ -1357,10 +1503,16 @@ public class Stub {
 		if (method.IsFinal) modifiers += " sealed";
 		if (modifiers == " virtual sealed") modifiers = "";
 	
+		MemberInfo[] defs = property.DeclaringType.GetDefaultMembers ();
 		string name = property.Name;
+		foreach (MemberInfo mi in defs) {
+			if (mi == property) {
+				name = "this";
+				break;
+			}
+		}
 	
-		string type_name = property.PropertyType.FullName;
-		type_name = ConvertCTSName (type_name);
+		string type_name = GetCSharpFullName (property.PropertyType);
 		
 		string parameters = GetMethodParameters (property.GetIndexParameters());
 		if (parameters != "") parameters = "[" + parameters + "]";		
@@ -1408,7 +1560,7 @@ public class Stub {
 		if (modifiers == " virtual sealed") modifiers = "";
 		
 		string name = ev.Name;
-		string type = ConvertCTSName(ev.EventHandlerType.FullName);
+		string type = GetCSharpFullName (ev.EventHandlerType);
 
 		return String.Format ("{0}{1} event {2} {3};",
 						visibility, modifiers, type, name);
@@ -1433,62 +1585,21 @@ public class Stub {
 	}
 
 	// Converts a fully .NET qualified type name into a C#-looking one
-	static string ConvertCTSName (string type) {
-		if (type.EndsWith ("[]"))
-			return ConvertCTSName(type.Substring(0, type.Length - 2).TrimEnd()) + "[]";
-
-		if (type.EndsWith ("&"))
-			return ConvertCTSName(type.Substring(0, type.Length - 1).TrimEnd()) + "&";
-
-		if (type.EndsWith ("*"))
-			return ConvertCTSName(type.Substring(0, type.Length - 1).TrimEnd()) + "*";
-
-		if (!type.StartsWith ("System."))
-				return type;
-		
-		switch (type) {
-		case "System.Byte": return "byte";
-		case "System.SByte": return "sbyte";
-		case "System.Int16": return "short";
-		case "System.Int32": return "int";
-		case "System.Int64": return "long";
-
-		case "System.UInt16": return "ushort";
-		case "System.UInt32": return "uint";
-		case "System.UInt64": return "ulong";
-
-		case "System.Single":  return "float";
-		case "System.Double":  return "double";
-		case "System.Decimal": return "decimal";
-		case "System.Boolean": return "bool";
-		case "System.Char":    return "char";
-		case "System.Void":    return "void";
-		case "System.String":  return "string";
-		case "System.Object":  return "object";
-		}
-		
-		// Types in the system namespace just get their type name returned.
-		if (type.StartsWith("System.")) {
-			string sysname = type.Substring(7);
-			if (sysname.IndexOf(".") == -1)
-				return sysname;
-		}
-
-		return type;
-	}
-	
 	private static string MakeSlashDocMemberSig(MemberInfo m) {
-		string typename = m.DeclaringType.FullName;
+		string typename = GetCSharpFullName (m.DeclaringType);
 		string memname = m.Name;
 		
-		if (m is ConstructorInfo)
+		if (m is ConstructorInfo) {
 			return "C:" + typename + ".#ctor(" + MakeSlashDocMemberSigArgs(((ConstructorInfo)m).GetParameters()) + ")";
-		if (m is MethodInfo)
+		}
+		if (m is MethodInfo) {
 			return "M:" + typename + "." + memname + MakeSlashDocMemberSigArgs(((MethodInfo)m).GetParameters()) + MakeSlashDocMemberSigReturn((MethodInfo)m);
+		}
 		if (m is PropertyInfo && ((PropertyInfo)m).GetIndexParameters().Length == 0)
 			return "P:" + typename + "." + memname;
-		if (m is PropertyInfo && ((PropertyInfo)m).GetIndexParameters().Length > 0)
+		if (m is PropertyInfo && ((PropertyInfo)m).GetIndexParameters().Length > 0) {
 			return "P:" + typename + "." + memname + MakeSlashDocMemberSigArgs(((PropertyInfo)m).GetIndexParameters());
+		}
 		if (m is FieldInfo)
 			return "F:" + typename + "." + memname;
 		if (m is EventInfo)
@@ -1498,18 +1609,149 @@ public class Stub {
 	
 	private static string MakeSlashDocMemberSigReturn(MethodInfo m) {
 		if (m.Name != "op_Implicit" && m.Name != "op_Explicit") return "";
-		return "~" + m.ReturnType.FullName;
+		return "~" + GetCSharpFullName (m.ReturnType);
 	}
 	
 	private static string MakeSlashDocMemberSigArgs(ParameterInfo[] ps) {
-		string ret = "";
+		StringBuilder buf = new StringBuilder ();
 		foreach (ParameterInfo p in ps) {
-			if (ret != "") ret += ",";
-			ret += p.ParameterType.FullName.Replace("&", "@");
+			if (buf.Length > 0)
+				buf.Append (",");
+			FillDocTypeName (p.ParameterType, buf, true, true, '+', true);
 		}
-		if (ret != "")
-			ret = "(" + ret + ")";
-		return ret;
+		if (buf.Length > 0)
+			return "(" + buf.ToString () + ")";
+		return "";
 	}
 
+	private static StringBuilder FillDocTypeName (Type type, StringBuilder buf, bool includeNamespace, bool includeDeclType, char declSep, bool csharp)
+	{
+		if (type == null)
+			throw new ArgumentNullException ("type");
+		if (type.IsArray) {
+			FillDocTypeName (type.GetElementType(), buf, includeNamespace, 
+				includeDeclType, declSep, csharp).Append ("[");
+			int rank = type.GetArrayRank ();
+			if (rank > 1)
+				buf.Append (new string (',', rank-1));
+			return buf.Append ("]");
+		}
+		if (type.IsByRef) {
+			return FillDocTypeName (type.GetElementType(), buf, 
+				includeNamespace, includeDeclType, declSep, csharp).Append ("@");
+		}
+		if (type.IsPointer) {
+			return FillDocTypeName (type.GetElementType(), buf, 
+				includeNamespace, includeDeclType, declSep, csharp).Append ("*");
+		}
+		if (type.IsGenericParameter)
+			return buf.Append (type.Name);
+		if (!type.IsGenericType) {
+			string t = type.FullName;
+			if (!csharp || !t.StartsWith ("System.")) {
+				if (includeNamespace)
+					buf.Append (type.Namespace).Append (".");
+				if (type.DeclaringType != null && includeDeclType)
+					buf.Append (type.DeclaringType.Name).Append (declSep);
+				return buf.Append (type.Name);
+			}
+			
+			switch (t) {
+			case "System.Byte":    return buf.Append ("byte");
+			case "System.SByte":   return buf.Append ("sbyte");
+			case "System.Int16":   return buf.Append ("short");
+			case "System.Int32":   return buf.Append ("int");
+			case "System.Int64":   return buf.Append ("long");
+
+			case "System.UInt16":  return buf.Append ("ushort");
+			case "System.UInt32":  return buf.Append ("uint");
+			case "System.UInt64":  return buf.Append ("ulong");
+
+			case "System.Single":  return buf.Append ("float");
+			case "System.Double":  return buf.Append ("double");
+			case "System.Decimal": return buf.Append ("decimal");
+			case "System.Boolean": return buf.Append ("bool");
+			case "System.Char":    return buf.Append ("char");
+			case "System.Void":    return buf.Append ("void");
+			case "System.String":  return buf.Append ("string");
+			case "System.Object":  return buf.Append ("object");
+			}
+			
+			// Types in the system namespace just get their type name returned.
+			if (includeNamespace && t.StartsWith("System.")) {
+				string sysname = t.Substring(7);
+				if (sysname.IndexOf(".") == -1)
+					return buf.Append (sysname);
+			}
+
+			return buf.Append (includeNamespace ? t : type.Name);
+		}
+		if (includeNamespace)
+			buf.Append (type.Namespace).Append (".");
+		Type[] genArgs = type.GetGenericArguments ();
+		int genArg = 0;
+		if (type.DeclaringType != null) {
+			if (includeDeclType) {
+				buf.Append (GetTypeName (type.DeclaringType.Name));
+				if (type.DeclaringType.IsGenericType) {
+					buf.Append ("<");
+					int max = type.DeclaringType.GetGenericArguments ().Length;
+					FillDocTypeName (genArgs [genArg++], buf, includeNamespace, 
+							includeDeclType, declSep, csharp);
+					while (genArg < max) {
+						buf.Append (",");
+						FillDocTypeName (genArgs [genArg++], buf, includeNamespace, 
+								includeDeclType, declSep, csharp);
+					}
+					buf.Append (">");
+				}
+				buf.Append (declSep);
+			}
+			else
+				genArg = type.DeclaringType.IsGenericType 
+					? type.DeclaringType.GetGenericArguments ().Length
+					: 0;
+		}
+		buf.Append (GetTypeName (type.Name));
+		if (genArg < genArgs.Length) {
+			buf.Append ("<");
+			FillDocTypeName (genArgs [genArg++], buf, includeNamespace, 
+					includeDeclType, declSep, csharp);
+			while (genArg < genArgs.Length) {
+				buf.Append (",");
+				FillDocTypeName (genArgs [genArg++], buf, includeNamespace, 
+						includeDeclType, declSep, csharp);
+			}
+			buf.Append (">");
+		}
+		return buf;
+	}
+
+	private static string GetDocTypeName (Type type)
+	{
+		return FillDocTypeName (type, new StringBuilder (), false, true, '+', false).ToString ();
+	}
+
+	private static string GetDocTypeFullName (Type type)
+	{
+		return FillDocTypeName (type, new StringBuilder (), true, true, '+', false).ToString ();
+	}
+
+	private static string GetCSharpName (Type type)
+	{
+		return FillDocTypeName (type, new StringBuilder (), false, true, '.', true).ToString ();
+	}
+
+	private static string GetCSharpFullName (Type type)
+	{
+		return FillDocTypeName (type, new StringBuilder (), true, true, '.', true).ToString ();
+	}
+
+	private static string GetTypeName (string name)
+	{
+		int n = name.IndexOf ("`");
+		if (n >= 0)
+			return name.Substring (0, n);
+		return name;
+	}
 }
